/* Libtlen2 - multiplatform Tlen.pl library
 * Copyright (C) 2003  Libtlen Team
 *
 * $Id: libtlen2-cmd.c,v 1.19 2004/05/29 19:16:04 who_ami Exp $
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */


#include "libtlen2-cmd.h"
#include "libtlen2-utils.h"

/**
  * tlen_notify:
  * @tc: connection to use.
  * @destination: recipient of this notify.
  * @type: type of notify
  *
  *
  * Send notify about typing, not typing or soundalert to a user.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/

gboolean tlen_notify (TlenConnection *tc, const gchar *destination,
                      TlenNotifyType type)
{
  gchar *query;
  gchar *notifytype;
  gboolean ret;
  gsize b;

  if (!tc)
  {
    tlen_debug ("tc = NULL\n");
    return FALSE;
  }
  if (!destination)
  {
    tlen_debug ("destination == NULL");
    return FALSE;
  }
  
  switch (type)
  {
    case TLEN_NOTIFY_TYPING:
      notifytype = "t";
      break;
    case TLEN_NOTIFY_STOPPED:
      notifytype = "u";
      break;
    case TLEN_NOTIFY_ALERT:
    default:
      notifytype = "a";
      break;
  }

  tlen_debug ("To: %s\nType: %s\n", destination, notifytype);
  query = g_strdup_printf ("<m t='%s' tp='%s'/>",
                           destination, notifytype);
  ret = tlen_connection_socket_writen (tc, query, tlen_strlen (query), &b);
  g_free (query);
  return ret;
}

/**
  * tlen_message:
  * @tc: connection to use.
  * @destination: recipient of this message.
  * @message: body of the message.
  * @type: type of the message.
  *
  *
  * Send message to a user.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/
gboolean tlen_message (TlenConnection *tc, const gchar *destination,
                       const gchar *message, TlenMessageType type)
{
  gchar *query;
  gchar *urlencoded;
  gchar *msgtype;
  gboolean ret;
  gsize b;

  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return FALSE;
  }
  if (!destination)
  {
    tlen_debug ("destination == NULL");
    return FALSE;
  }
  if (!message)
  {
    tlen_debug ("message == NULL");
    return FALSE;
  }
  urlencoded = tlen_encode (message);

  if (!urlencoded)
  {
    tlen_debug ("Error encoding the message.\n");
    return FALSE;
  }
  switch (type)
  {
    case TLEN_MESSAGE_TYPE_NORMAL:
      msgtype = "normal";
      break;
    case TLEN_MESSAGE_TYPE_CHAT:
      msgtype = "chat";
      break;
    default:
      msgtype = "chat";
      break;
  }

  tlen_debug ("To: %s\nMessage: %s\nTyp: %s\n", destination, message, msgtype);
  query = g_strdup_printf ("<message to='%s' type='%s'><body>%s</body></message>",
                           destination, msgtype, urlencoded);
  ret = tlen_connection_socket_writen (tc, query, tlen_strlen (query), &b);
  g_free (query);
  g_free (urlencoded);
  return ret;
}

/**
  * tlen_roster_get:
  * @tc: connection to use.
  *
  * Get roster from server.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/
gboolean tlen_roster_get (TlenConnection *tc)
{
  gboolean ret;
  gchar *q;
  gsize b;

  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }

  if (tc->state != TLEN_STATE_LOGGED)
    return FALSE;
  q = "<iq type='get' id='GetRoster'><query xmlns='jabber:iq:roster'></query></iq>";
  ret = tlen_connection_socket_writen (tc, q, tlen_strlen(q), &b);

  return ret;
}

/**
  * tlen_presence:
  * @tc: connection to use.
  * @status: what status to send.
  * @description: description of status, maybe NULL.
  *
  * Change user status on server.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/
gboolean tlen_presence (TlenConnection *tc, TlenPresenceType status,
                        const gchar *description)
{
  gchar *presence;
  gchar *query;
  gchar *urlencoded;
  gboolean ret = TRUE;
  gsize bytes;

  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }

  if (tc->state != TLEN_STATE_LOGGED)
    return FALSE;

  urlencoded = tlen_encode (description ? description : "");
  tc->status = status;
  g_free (tc->description);
  tc->description = g_strdup (description);

  switch (status)
  {
    case TLEN_PRESENCE_AVAILABLE:
      presence = "available";
      break;
    case TLEN_PRESENCE_AWAY:
      presence = "away";
      break;
    case TLEN_PRESENCE_CHATTY:
      presence = "chat";
      break;
    case TLEN_PRESENCE_EXT_AWAY:
      presence = "xa";
      break;
    case TLEN_PRESENCE_DND:
      presence = "dnd";
      break;
    case TLEN_PRESENCE_INVISIBLE:
      query = g_strdup_printf ("<presence type='invisible'><status>%s</status></presence>", urlencoded);
      ret = tlen_connection_socket_writen (tc, query, tlen_strlen (query),
                                           &bytes);
      query = NULL;
      return ret;
      break;
    case TLEN_PRESENCE_UNAVAILABLE:
      query = g_strdup_printf ("<presence type='unavailable'><status>%s</status></presence>", urlencoded);
      ret = tlen_connection_socket_writen (tc, query, tlen_strlen (query),
                                           &bytes);
      if (!ret)
        return FALSE;
      query = "</s>";
      ret = tlen_connection_socket_writen (tc, query, tlen_strlen (query),
                                           &bytes);
      query = NULL;
      tc->state = TLEN_STATE_UNAVAILABLE;
      return ret;
      break;
    default:
      presence = "available";
      break;
  }

  tlen_debug ("Status: %s\nDescription: %s\n", presence, description);

  query = g_strdup_printf ("<presence><show>%s</show><status>%s</status></presence>", presence, urlencoded);
  ret = tlen_connection_socket_writen (tc, query, tlen_strlen (query),
                                       &bytes);
  g_free (query);
  g_free (urlencoded);

  return ret;
}

/**
  * tlen_ping:
  * @data: #TlenConnection to use, specified in gpointer
  *
  * Sends a keep-alive packet to the server. Must be run on every minute.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/
gboolean _tlen_ping (gpointer data)//todo: raise event disconnected?
{
  gsize bytes;
  gchar *q;

  if (!(TlenConnection*)data)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }

  tlen_debug_high ("Keepalive sent\n");
  q = "  \t  ";
  return tlen_connection_socket_writen ((TlenConnection*)data, q,
                                        tlen_strlen (q), &bytes);
}
