/* Libtlen2 - multiplatform Tlen.pl library
 * Copyright (C) 2003  Libtlen Team
 *
 * $Id: libtlen2-connection.c,v 1.23 2004/02/28 13:12:43 who_ami Exp $
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */


#include "libtlen2.h"
#include "libtlen2-parser.h"
#include "libtlen2-events.h"
#include "libtlen2-utils.h"
#include "libtlen2-login.h"
#include <string.h>

#define HTTP_MAX_RESPONSE_LEN 8192

static
void _tlen_hub_async_connfunc (GTcpSocket *socket,
                               GTcpSocketConnectAsyncStatus status,
                               gpointer data);
static
gboolean _tlen_hub_async_iofunc (GIOChannel* iochannel, GIOCondition condition,
                                 gpointer data);
static
void _tlen_use_fallback_server (TlenConnection *tc);
static
void _tlen_async_connfunc (GTcpSocket *socket,
                           GTcpSocketConnectAsyncStatus status,
                           gpointer data);
static
gboolean _tlen_async_iofunc (GIOChannel* iochannel, GIOCondition condition,
                             gpointer data);
static
void _tlen_server_async_connect (TlenConnection *tc);

/**
  * tlen_connection_new:
  * @username: user used while logging into the server.
  * @password: used password.
  *
  * Creates new TlenConnection object with given username and password.
  *
  * Return value: new #TlenConnection on success, NULL otherwise.
  **/
TlenConnection*
tlen_connection_new(const gchar *username,
                    const gchar *password)
{
  TlenConnection *tc = NULL;

  tc = g_new0 (TlenConnection, 1);
  if (tc)
  {
    tc->username = g_strdup(username);
    tc->password = g_strdup(password);
  }
  return tc;
}

/**
  * tlen_connection_open:
  * @tc: pointer to #TlenConnection.
  *
  * Opens new connection with the server in non-blocking mode.
  *
  * Return value: TRUE on success, FALSE otherwise.
  **/
gboolean
tlen_connection_open(TlenConnection *tc, TlenEventCallback event_cb)
{
  if (!tc)
  {
    tlen_debug ("Assertion tc != NULL failed.\n");
    return FALSE;
  }
  tc->event_cb = event_cb;
  tc->state = TLEN_STATE_CONNECTING_HUB;
  tlen_debug ("Connecting to %s:80...\n", TLEN_HUB);
  tc->socket = gnet_tcp_socket_connect_async (TLEN_HUB, 80,
                                              _tlen_hub_async_connfunc,
                                              (gpointer) tc);

  return TRUE;
}

/**
  * tlen_connection_close:
  * @tc: #TlenConnection to close.
  *
  * Closes given connection with the server by sending presence of type
  * unavailable.
  *
  * Note: You should wait for TLEN_EVENT_DISCONNECTED before you run
  * #tlen_connection_delete.
  *
  * Return value: TRUE on success, FALSE otherwise.
  *
  **/
gboolean
tlen_connection_close(TlenConnection *tc)
{
  if (!tc)
  {
    tlen_debug ("Assertion tc != NULL failed.\n");
    return FALSE;
  }
  if (tc->state == TLEN_STATE_LOGGED)
    tlen_presence (tc, TLEN_PRESENCE_UNAVAILABLE, NULL);
  else if ((tc->state == TLEN_STATE_CONNECTING_HUB) ||
           (tc->state == TLEN_STATE_CONNECTING))
  {
    tc->state = TLEN_STATE_NOT_CONNECTED;
    tlen_debug ("Cancelling connection\n");
    gnet_tcp_socket_connect_async_cancel (tc->socket);
    tc->socket = NULL;
  }
  return TRUE;
}

/**
  * tlen_connection_delete:
  * @tc: #TlenConnection to delete.
  *
  * Frees object pointed by @tc. If there are opened connections associated
  * with this object then they are closed.
  *
  * Note: You should run #tlen_connection_close first if your were logged.
  *
  * Return value: TRUE on success, FALSE otherwise.
  *
  **/
gboolean
tlen_connection_delete(TlenConnection *tc)
{
  if(!tc)
  {
    tlen_debug ("Assertion tc != NULL failed.\n");
    return FALSE;
  }
  if (tc->state == TLEN_STATE_LOGGED)
  {
    tlen_debug ("Still logged, wait for TLEN_EVENT_DISCONNECTED or run tlen_connection_close first.\n");
    return FALSE;
  }
  if (tc->socket)
  {
    tlen_debug ("Destroying connection socket...\n");
    gnet_tcp_socket_delete (tc->socket);
    tc->socket = NULL;
  }
  g_free (tc->username);
  g_free (tc->password);
  g_free (tc->hostname);
  g_free (tc->description);
  g_free (tc->sid);
  if (tc->buffer)
    g_string_free (tc->buffer, TRUE);
  g_free (tc);
  return TRUE;
}

/**
  * tlen_connection_socket_writen:
  * @tc: #TlenConnection object to use.
  * @buffer: buffer to read from.
  * @length: length of @buffer.
  * @bytes_n: pointer to integer in which to store the number of bytes written.
  *
  * Writes all length bytes in buffer to @tc.
  * If @bytes_n is set, the number of bytes written is stored in the integer
  * it points to. bytes_n will be less than length if the connection closed
  * or an error occured.
  *
  * Return value: TRUE on success, FALSE otherwise.
  *
  **/
gboolean
tlen_connection_socket_writen (TlenConnection* tc, gpointer buffer,
                               gsize length, gsize* bytes_n)
{
  GIOError err;
  GIOChannel* iochannel;

  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }
  if (!buffer)
  {
    tlen_debug ("buffer == NULL!\n");
    return FALSE;
  }
  if (tc->state < TLEN_STATE_CONNECTED_TO_HUB)
  {
    tlen_debug ("Not connected to server or hub!\n");
    return FALSE;
  }
  iochannel = gnet_tcp_socket_get_io_channel (tc->socket);
  if (!iochannel)
  {
    tlen_debug ("no iochannel available for this socket.\n");
    return FALSE;
  }
  tlen_debug_high ("sending to server:'%s'\n", (gchar*) buffer);
  err = gnet_io_channel_writen (iochannel, buffer, length, bytes_n);
  if ((err != G_IO_ERROR_NONE) || (length != *bytes_n))
  {
    tlen_debug ("Error occured while writing to socket.\n");
    return FALSE;
  }
  return TRUE;
}

/**
  * tlen_connection_socket_readn:
  * @tc: #TlenConnection object to use.
  * @buffer: buffer to write to.
  *
  * Read @length bytes from @tc into @buffer.  If @bytes_n is set, the number
  * of bytes read is stored in the integer it points to. @bytes_n will be less
  * than @length if the end-of-file was reached or an error occured.
  *
  * Return value: TRUE on success, FALSE otherwise.
  *
  **/
gboolean
tlen_connection_socket_readn (TlenConnection* tc, gpointer buffer,
                              gsize length, gsize* bytes_n)
{
  GIOError err;
  GIOChannel* iochannel = NULL;

  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }
  if (tc->state < TLEN_STATE_CONNECTED_TO_HUB)
  {
    tlen_debug ("Not connected to server or hub!\n");
    return FALSE;
  }
  iochannel = gnet_tcp_socket_get_io_channel (tc->socket);
  if (!iochannel)
  {
    tlen_debug ("no iochannel available for this socket.\n");
    return FALSE;
  }
  err = gnet_io_channel_readn (iochannel, buffer, length,  bytes_n);
  if (err != G_IO_ERROR_NONE)
  {
    tlen_debug ("Error occured while reading from socket.\n");
    return FALSE;
  }
//  tlen_debug_high ("read from server:'%s'\n", buffer);
  return TRUE;
}

/**
  * tlen_connection_socket_readline:
  * @tc: #TlenConnection object to use.
  * @buffer: pointer to gchar* in which to store the new buffer.
  *
  * Read a line from @tc. @buffer must be freed after use.
  *
  * Return value: TRUE on success, FALSE otherwise.
  *
  **/
gboolean
tlen_connection_socket_readline (TlenConnection* tc, gchar** buffer)
{
  gsize bytes_n;
  GIOError err;
  GIOChannel* iochannel = NULL;

  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }
  if (tc->state < TLEN_STATE_CONNECTED_TO_HUB)
  {
    tlen_debug ("Not connected to server or hub!\n");
    return FALSE;
  }
  iochannel = gnet_tcp_socket_get_io_channel (tc->socket);
  if (!iochannel)
  {
    tlen_debug ("no iochannel available for this socket.\n");
    return FALSE;
  }
  err = gnet_io_channel_readline_strdup (iochannel, buffer, &bytes_n);
  if (err != G_IO_ERROR_NONE)
  {
    tlen_debug ("Error occured while reading from socket.\n");
    return FALSE;
  }
  tlen_debug_high ("read from server:'%s'\n", buffer);
  return TRUE;
}


static
void _tlen_hub_async_connfunc (GTcpSocket *socket,
                               GTcpSocketConnectAsyncStatus status,
                               gpointer data)
{
  TlenConnection *tc;
  GIOChannel *iochannel;
  gchar *q;
  gboolean ret;
  gsize len;

  tc = (TlenConnection *) data;
  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return;
  }
  if (status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK)
  {
    tlen_debug ("Error: Could not connect to HUB.\n");
    _tlen_use_fallback_server (tc);
  }
  else
  {
    tlen_debug ("Connected.\n");
    tc->state = TLEN_STATE_CONNECTED_TO_HUB;
    tc->socket = socket;
    iochannel = gnet_tcp_socket_get_io_channel (tc->socket);
    if (!iochannel)
    {
      tlen_debug ("Error getting IOchannel.\n");
      _tlen_use_fallback_server (tc);
      return;
    }
    g_io_add_watch (iochannel, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
                    _tlen_hub_async_iofunc, (gpointer) tc);
    q = g_strdup_printf ("GET /4starters.php?u=%s&v=10 HTTP/1.0\r\nHost: idi.tlen.pl\r\n\r\n",
                         tc->username);
    tlen_debug ("Sending query to HUB...\n");
    ret = tlen_connection_socket_writen (tc, q, tlen_strlen (q), &len);
    g_free (q);
  }
}

static
void _tlen_use_fallback_server (TlenConnection *tc)
{
//  TlenEvent ev;

  if (tc->state != TLEN_STATE_NOT_CONNECTED);
  {
    tlen_debug ("Using fallback server.\n");
    tc->hostname = g_strdup(TLEN_FALLBACK_SERVER);
    tc->port = TLEN_FALLBACK_PORT;
    _tlen_server_async_connect (tc);
  }
//  else // connection was cancelled
//  {
//    ev = tlen_event_new (TLEN_EVENT_DISCONNECTED);
//    tlen_event_add (tc, ev);
//  }
}

static
gboolean _tlen_hub_async_iofunc (GIOChannel* iochannel, GIOCondition condition,
                                 gpointer data)
{
  gboolean ret;
  gchar response[HTTP_MAX_RESPONSE_LEN+1];
  gsize len;
  gchar *body;
  gchar *server;
  guint port;
  TlenConnection *tc;

  tc = (TlenConnection*) data;
  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }
  /* Check for socket error */
  if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
  {
    tlen_connection_close (tc);
    _tlen_use_fallback_server (tc);
    return FALSE;
  }

  /* Check for data to be read */
  if (condition & G_IO_IN)
  {
    tlen_debug ("Reading HUB response...\n");
    ret = tlen_connection_socket_readn (tc, (gpointer) response,
                                        HTTP_MAX_RESPONSE_LEN, &len);
    if (!ret || (len == 0))
    {
      tlen_debug ("Disconnecting from HUB...\n");
      tlen_connection_close (tc);
      _tlen_use_fallback_server (tc);
      return FALSE;
    }
    response[len] = '\0';
    tlen_debug_high ("HUB response: '%s'\n", response);
    body = strstr (response,"\r\n\r\n");
    if (!body)
      body = ""; /* If somehow it was NULL, parser would crash. */
    ret = _tlen_parse_hub_response (tc, body, &server, &port);
    if (!ret)
    {
      tlen_debug ("Disconnecting from HUB...\n");
      tlen_connection_close (tc);
      _tlen_use_fallback_server (tc);
    }
    else
    {
      tlen_debug ("Using server: %s:%d.\n", server, port);
      tc->hostname = server;
      tc->port = port;
      tlen_debug ("Disconnecting from HUB...\n");
      tlen_connection_close (tc);
      _tlen_server_async_connect (tc);
    }
    return FALSE;
  }
  return TRUE;
}

static
void _tlen_server_async_connect (TlenConnection *tc)
{
  tc->state = TLEN_STATE_CONNECTING;
  tlen_debug ("Connecting to %s:%d...\n", tc->hostname, tc->port);
  tc->socket = gnet_tcp_socket_connect_async (tc->hostname, tc->port,
                                              _tlen_async_connfunc,
                                              (gpointer) tc);
}

static
void _tlen_async_connfunc (GTcpSocket *socket,
                           GTcpSocketConnectAsyncStatus status,
                           gpointer data)
{
  TlenEvent *ev;
  TlenConnection *tc;
  GIOChannel *iochannel;

  tc = (TlenConnection *) data;
  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return;
  }
  if (status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK)
  {
    tlen_debug ("Error: Could not connect (status = %d)\n", status);
    tc->state = TLEN_STATE_NOT_CONNECTED;
    ev = tlen_event_new (TLEN_EVENT_CONNECTION_FAILED);
    tlen_event_add (tc, ev);
    return;
  }
  tc->socket = socket;
  tlen_debug ("Connected.\n");
  tc->state = TLEN_STATE_CONNECTED;
  ev = tlen_event_new (TLEN_EVENT_CONNECTED);
  tlen_event_add (tc, ev);
  iochannel = gnet_tcp_socket_get_io_channel (tc->socket);
  if (!iochannel)
  {
    tlen_debug ("Error getting IOchannel.\n");
    tc->state = TLEN_STATE_NOT_CONNECTED;
    ev = tlen_event_new (TLEN_EVENT_CONNECTION_FAILED);
    tlen_event_add (tc, ev);
    return;
  }
  g_io_add_watch (iochannel, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
                  _tlen_async_iofunc, (gpointer) tc);
  _tlen_login (tc);
}

static
gboolean _tlen_async_iofunc (GIOChannel* iochannel, GIOCondition condition,
                             gpointer data)
{
  TlenConnection *tc;
  TlenEvent *ev;

  tc = (TlenConnection*) data;
  if (!tc)
  {
    tlen_debug ("tc == NULL!\n");
    return FALSE;
  }
  /* Check for socket error */
  if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
  {
    tc->state = TLEN_STATE_NOT_CONNECTED;
    ev = tlen_event_new (TLEN_EVENT_DISCONNECTED);
    tlen_event_add (tc, ev);
    return FALSE;
  }

  /* Check for data to be read */
  if (condition & G_IO_IN)
  {
    GError *error = NULL;
    GIOStatus status;
    gchar buffer[TLEN_IO_READ_BUFFER_SIZE];
    guint bytes_read;

    tlen_debug ("data to be read\n");
    /* Read the data into our buffer */
    status = g_io_channel_read_chars (iochannel, buffer,
                                      sizeof(buffer)-1, &bytes_read, &error);

    /* Check for stdin error */
    if (status !=  G_IO_STATUS_NORMAL)
    {
      tc->state = TLEN_STATE_NOT_CONNECTED;
      ev = tlen_event_new (TLEN_EVENT_DISCONNECTED);
      tlen_event_add (tc, ev);
      return FALSE;
    }

    /* Check for EOF */
    else if (bytes_read == 0)
    {
      tlen_debug ("EOF on connection.\n");
      tc->state = TLEN_STATE_NOT_CONNECTED;
      ev = tlen_event_new (TLEN_EVENT_DISCONNECTED);
      tlen_event_add (tc, ev);
      return FALSE;
    }

    /* Otherwise, print */
    else
    {
      buffer [bytes_read] = '\0';
      tlen_debug_high ("data read: '%s'\n", buffer);
      if (_tlen_parser_buffer_append (tc, buffer) == FALSE)
      {
        /* FIXME: Not so easy - what about the status? */
        return FALSE;
      }
    }
  }

  return TRUE;
}
