/* Libtlen2 - multiplatform Tlen.pl library
 * Copyright (C) 2003  Libtlen Team
 *
 * $Id: libtlen2-p2p.c,v 1.4 2004/03/09 12:56:27 who_ami Exp $
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */



#include "libtlen2-p2p.h"
#include "libtlen2-events.h"

#include <string.h>
#include <time.h>
#ifdef __WIN32__
#include <sys/types.h>
#endif

static GSList *list_p2p;

static void
_p2p_async_accept (GTcpSocket* server, GTcpSocket* client, gpointer data);
static void
_p2p_async_connfunc (GTcpSocket* socket,
                     GTcpSocketConnectAsyncStatus status, gpointer data);

gboolean tlen_p2p_connect (TlenConnection *tc, const gchar *hostname,
                           gint port, const gchar *i)
{
  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return FALSE;
  }
  if (!hostname)
  {
    tlen_debug ("hostname == NULL\n");
    return FALSE;
  }
  if (!i)
  {
    tlen_debug ("i == NULL\n");
    return FALSE;
  }
  tlen_debug ("Connecting with %s:%d\n", hostname, port);
  _tlen_p2p_add (g_strdup (i));
  gnet_tcp_socket_connect_async (hostname, port, _p2p_async_connfunc,
                                 tc);
  return TRUE;
}

gboolean tlen_p2p_open (TlenConnection *tc, const gchar *id,
                        const gchar *i)
{
  gchar *q, *s;
  gsize bytes;
  gboolean ret;
  GTcpSocket *server;
  GInetAddr *addr = NULL, *addr2;
  gchar *name = NULL;
  gint port = 0;
  gboolean error = FALSE;

  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return FALSE;
  }
  if (!id)
  {
    tlen_debug ("id == NULL\n");
    return FALSE;
  }
  if (!i)
  {
    tlen_debug ("i == NULL\n");
    return FALSE;
  }
  s = strchr (id, '@');
  if (s)
    *s = '\0';
  server = gnet_tcp_socket_server_new_with_port (port);
  if (!server)
  {
    tlen_debug ("Could not create p2p server on port %d\n", port);
    error = TRUE;
  }
  else
  {
    addr = gnet_tcp_socket_get_local_inetaddr (server);
    addr2 = gnet_inetaddr_autodetect_internet_interface ();
    if (!addr || !addr2)
    {
      tlen_debug ("Could not get inetaddr\n");
      gnet_tcp_socket_delete (server);
      error = TRUE;
    }
    else
    {
      name = gnet_inetaddr_get_canonical_name (addr2);
      port = gnet_inetaddr_get_port (addr);
      tlen_debug ("Running p2p server on port %d\n", port);
      gnet_inetaddr_delete (addr);
      gnet_inetaddr_delete (addr2);
    }
  }
  if (error)
    q = g_strdup_printf ("<f t=\"%s\" e=\"%d\" i=\"%s\"/>",
                         id, TLEN_P2P_ERROR, i);
  else
  {
    q = g_strdup_printf ("<f t=\"%s\" e=\"%d\" i=\"%s\" a=\"%s\" p=\"%d\"/>",
                         id, TLEN_P2P_OPENED, i, name, port);
    _tlen_p2p_add (g_strdup (i));
    gnet_tcp_socket_server_accept_async (server, _p2p_async_accept, tc);
  }

  ret = tlen_connection_socket_writen (tc, q, tlen_strlen (q), &bytes);
  g_free (q);
  if (s)
    *s = '@';
  return ret;
}

/**
  * tlen_p2p_denyfile:
  * @tc: connection to use.
  * @id: id of user.
  * @i: id of p2p session.
  *
  * Deny a file.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/
gboolean tlen_p2p_denyfile (TlenConnection *tc, const gchar *id,
                            const gchar *i)
{
  gchar *q, *s;
  gsize bytes;
  gboolean ret;

  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return FALSE;
  }
  if (!id)
  {
    tlen_debug ("id == NULL\n");
    return FALSE;
  }
  if (!i)
  {
    tlen_debug ("i == NULL\n");
    return FALSE;
  }
  s = strchr (id, '@');
  if (s)
    *s = '\0';
  q = g_strdup_printf ("<f t=\"%s\" e=\"%d\" i=\"%s\" v=\"1\"/>",
                       id, TLEN_P2P_DENIED, i);
  ret = tlen_connection_socket_writen (tc, q, tlen_strlen (q), &bytes);
  g_free (q);
  if (s)
    *s = '@';
  return ret;
}

/**
  * tlen_p2p_sendfile:
  * @tc: connection to use.
  * @id: receipent's id.
  * @filename: name of a file.
  * @i: pointer to newly allocated string containing id of this p2p session.
  *
  * Send file to a user using P2P connection.
  *
  * returns: TRUE on success, FALSE otherwise.
  **/
gboolean tlen_p2p_sendfile (TlenConnection *tc, const gchar *id,
                            const gchar *filename, gchar **i)
{
  TlenP2p *p;
  gchar *q, *s;
  gsize bytes;
  gboolean ret;
  time_t t;
  struct tm *tm;
  gchar s1[7];

  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return FALSE;
  }
  if (!id)
  {
    tlen_debug ("id == NULL\n");
    return FALSE;
  }
  if (!filename)
  {
    tlen_debug ("filename == NULL\n");
    return FALSE;
  }
  s = strchr (id, '@');
  if (s)
    *s = '\0';
  p = _tlen_p2p_new (id);
  if (!p)
  {
    if (s)
      *s = '@';
    return FALSE;
  }
  t = time (NULL);
  tm = localtime (&t);
  strftime (s1, 11, "%H%M%S", tm);
  p->i = g_strdup(s1);
  q = g_strdup_printf ("<f t=\"%s\" n=\"%s\" e=\"1\" i=\"%s\" c=\"1\" s=\"%d\" v=\"1\"/>",
                       id, filename, s1, p->length);
  ret = tlen_connection_socket_writen (tc, q, tlen_strlen (q), &bytes);
  _tlen_p2p_free (p);
  if (ret)
    *i = g_strdup (s1);
  g_free (q);
  if (s)
    *s = '@';
  return ret;
}

void _tlen_p2p_add (gchar *i)
{
  list_p2p = g_slist_append (list_p2p, i);
}

void _tlen_p2p_del (gchar *i)
{
  list_p2p = g_slist_remove (list_p2p, i);
  g_free (i);
}

TlenP2p *_tlen_p2p_new (const gchar *id)
{
  TlenP2p *p;

  p = g_new0 (TlenP2p, 1);
  if (p)
  {
    p->id = g_strdup (id);
  }
  return p;
}

void _tlen_p2p_free (TlenP2p *p)
{
  if (p)
  {
    g_free (p->id);
    g_free (p->buf);
    g_free (p->filename);
    g_free (p->i);
    g_free (p->hostname);
    g_free (p);
  }
}

const gchar *_tlen_p2p_get (void)
{
  GSList *l;
  gchar *p;

  l = list_p2p;
  if (l)
  {
    p = (gchar*) l->data;
    return p;
  }
  return NULL;
}

static void
_p2p_async_accept (GTcpSocket* server, GTcpSocket* client, gpointer data)
{
  TlenConnection *tc = (TlenConnection*) data;

  if (client)
  {
    tlen_debug ("P2p connection accepted\n");
  }
  else
  {
    tlen_debug ("Error while accepting p2p connection\n");
  }
}

static void
_p2p_async_connfunc (GTcpSocket* socket,
                     GTcpSocketConnectAsyncStatus status, gpointer data)
{
  TlenEvent *ev;
  TlenConnection *tc = (TlenConnection*) data;
  gchar *i;

  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return;
  }
  ev = tlen_event_new (TLEN_EVENT_P2P);
  if (status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK)
  {
    tlen_debug ("Connection failed\n");
  }
  else
  {
    tlen_debug ("Connected\n");
    TLEN_EVENT_P2P(ev)->socket = socket;
    i = (gchar*) _tlen_p2p_get ();
    tlen_debug ("found i '%s'\n", i);
    TLEN_EVENT_P2P(ev)->i = g_strdup (i);
    _tlen_p2p_del (i);
  }
  tlen_event_add (tc, ev);
}


