/* Libtlen2 - multiplatform Tlen.pl library
 * Copyright (C) 2003  Libtlen Team
 *
 * $Id: libtlen2-parser.c,v 1.43 2004/05/29 19:16:04 who_ami Exp $
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include "libtlen2-parser.h"
#include "libtlen2-cmd.h"
#include "libtlen2-utils.h"
#include "libtlen2-events.h"
#include "libtlen2-login.h"
#include "libtlen2-p2p.h"
#include "libtlen2.h"
#include <string.h>
#include <stdlib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

static
void _tlen_parse_s (TlenConnection *tc, xmlNodePtr cur);
static
void _tlen_parse_iq (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
void _tlen_parse_presence (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
void _tlen_parse_message (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
void _tlen_parse_roster (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
void _tlen_parse_notify (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
TlenPubdirItem* _tlen_parse_tuba_item (TlenConnection *tc, xmlDocPtr doc,
                                       xmlNodePtr
                            cur);
static
void _tlen_parse_search_result (TlenConnection *tc, xmlDocPtr doc,
                                xmlNodePtr cur);
static
void _tlen_parse_newmail (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
void _tlen_parse_webmsg (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);
static
void _tlen_parse_file (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur);


static
gboolean _tlen_parser_buffer_parse (TlenConnection *tc, const gchar *s);
gchar *_tlen_parser_find_s(GString *gs);
gchar *_tlen_parser_find_valid_xml(GString *gs, gboolean *end_tag);

gboolean
_tlen_parse_hub_response (TlenConnection *tc, const gchar *body,
                          gchar **server, guint *port)
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  xmlChar *atr, *atr2;

  doc = xmlParseMemory (body, strlen (body));
  if (!doc)
  {
    tlen_debug ("Error parsing HUB response.\n");
    return FALSE;
  }
  cur = xmlDocGetRootElement(doc);
  if (!cur)
  {
    tlen_debug ("Empty document.\n");
    xmlFreeDoc(doc);
    return FALSE;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *) "t"))
  {
    tlen_debug ("Document of the wrong type, root node != t");
    xmlFreeDoc(doc);
    return FALSE;
  }
  atr = xmlGetProp(cur, "s");
  atr2 = xmlGetProp(cur, "p");
  if (!atr)
  {
    tlen_debug ("No s or p attribute in document.\n");
    xmlFreeDoc(doc);
    return FALSE;
  }
  *server = g_strdup ((gchar*) atr);
  *port = atoi ((const char*) atr2);
  xmlFreeDoc(doc);
  return TRUE;
}

/*-
  * _tlen_parser_find_s:
  * @s: a GString
  *
  * Checks if the GString contains a complete opening tag
  * e.g. <s i='33'>.
  *
  * Side effects: The found tag is cut away from gs.
  *
  * Return value: a string with the found tag + </s>.
  *               NULL if no complete opening tag has been found.
  *
  **/
gchar *_tlen_parser_find_s(GString *gs)
{
  gchar *rc         = NULL;
  gchar *s          = gs->str;
  gchar *found      = NULL;

  if (s[0] == '<')
  {
    found = strchr(&s[1], '>');
    if (found)
    {
      guint xml_len = (found - s) + 1;
      gchar *tmp;

      tmp = g_strndup(s, xml_len);
      rc = g_strconcat(tmp, "</s>", NULL); /* Yes - this could be optimized away. */
      g_free(tmp);
      g_string_erase(gs, 0, xml_len);
    }
  }
  else
  {
    /* Not possible unless server is high - what now? */
  }
  return rc;
}

/*-
  * _tlen_parser_find_valid_xml:
  * @s: a GString
  * @end_tag: end of XML marker
  *
  * Checks if the GString contains an XML document,
  * i.e. there is a substring starting from the begining of s
  * where the number of opening and closing tags is equal.
  *
  * Note: very simple - assumes that string *is* XML.
  * Note: the outermost tag cannot be a <sth/> style tag.
  *
  * Side effect: The found XML document is cut away from gs.
  * Side effect: The *end_tag variable is set to TRUE if
  *              the first thing in the GString is a closing tag.
  *
  * Return value: a string with the XML document;
  *               NULL if not all tags are closed or string is of zero length;
  *               NULL if end_tag is set to TRUE on exit.
  *
  **/
gchar *_tlen_parser_find_valid_xml(GString *gs, gboolean *end_tag)
{
  guint tags_opened = 0;
  guint len         = gs->len;
  gchar *rc         = NULL;
  guint i           = 0;
  gchar *s          = gs->str;

  if (len)
  {
    if ((len > 1) && (s[0] == '<') && (s[1] == '/'))
    {
      /* No kidding  - this must be </s>. */
      *end_tag = TRUE;
    }
    else if (s[0] == '<')
    {
      tags_opened = 1;
      for (i = 1; i < len; i++)
      {
        if (s[i] == '<')
        {
          if (((i+1) < len) && (s[i+1] == '/'))
          {
            tags_opened--;
          }
          else
          {
            tags_opened++;
          }
        }
        else if ((s[i] == '>') && (s[i-1] == '/'))
        {
          tags_opened--;
        }
        if (tags_opened == 0)
        {
          /* We've got the begining of the level 0 closing tag. */
          break;
        }
      } /* for */
      if (tags_opened == 0)
      {
        /* The number of tags matches.
           'i' points to '<' in the closing tag.
           Find the nearest '>'.
        */
        gchar *found = strchr(&s[i], '>');

        if (found)
        {
          guint xml_len = i + (found - &s[i]) + 1;

          rc = g_strndup(s, xml_len); /* What if it fails? */
          g_string_erase(gs, 0, xml_len);
        }
      } /* if */
    }
    else
    {
      /* Something is definetly wrong - string should start with a tag. */
      /* What should I do? */
    }
  }
  return rc;
}

gboolean
_tlen_parser_buffer_append(TlenConnection *tc, const gchar *s)
{
  gboolean rc = TRUE;

  if (!tc)
  {
    tlen_debug("tc == NULL!\n");
    rc = FALSE;
  }
  if (!s)
  {
    tlen_debug("s == NULL!\n");
    rc = FALSE;
  }

  if (rc)
  {
    if (tc->buffer)
    {
      tc->buffer = g_string_append(tc->buffer, s);
    }
    else
    {
      tc->buffer = g_string_new(s);
    }

    while (tc->buffer->len)
    {
      gchar    *xml    = NULL;
      gboolean end_tag = FALSE;

      if (!tc->id_parsed)
      {
        xml = _tlen_parser_find_s(tc->buffer);
      }
      else
      {
        xml = _tlen_parser_find_valid_xml(tc->buffer, &end_tag);
      }

      if (end_tag)
      {
        TlenEvent *ev;

        g_string_free(tc->buffer, TRUE);
        tc->buffer = NULL;
        tlen_debug("Server closed session.\n");
        ev = tlen_event_new(TLEN_EVENT_DISCONNECTED);
        tlen_event_add(tc, ev);
        break;
      }

      if (xml)
      {
        if (!tc->id_parsed) tc->id_parsed = TRUE;
        _tlen_parser_buffer_parse(tc, xml);
        g_free(xml);
        xml = NULL;
      }
      else
      {
        break;
      }
    }
  }
  return rc;
}

static
gboolean _tlen_parser_buffer_parse (TlenConnection *tc, const gchar *s)
{
  xmlDocPtr doc;
  xmlNodePtr cur;

  if (!tc)
  {
    tlen_debug ("tc == NULL\n");
    return FALSE;
  }
  if (!s)
  {
    tlen_debug ("s == NULL\n");
    return FALSE;
  }
//  tlen_debug_high ("Parser: '%s'\n", tc->buffer->str);

  doc = xmlParseMemory(s, tlen_strlen (s));

  if (doc == NULL )
  {
    tlen_debug ("Buffer not parsed successfully.\n");
    return FALSE;
  }

  cur = xmlDocGetRootElement(doc);

  if (cur == NULL)
  {
    tlen_debug ("Empty buffer.\n");
    xmlFreeDoc(doc);
    return FALSE;
  }

  if (!xmlStrcmp(cur->name, (const xmlChar *) "s"))
  {
    _tlen_parse_s (tc, cur);
  }
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "iq"))
  {
    _tlen_parse_iq (tc, doc, cur);
  }
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "presence"))
  {
    _tlen_parse_presence (tc, doc, cur);
  }
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "message"))
  {
    _tlen_parse_message (tc, doc, cur);
  }
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "n"))
    _tlen_parse_newmail (tc, doc, cur);
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "w"))
    _tlen_parse_webmsg (tc, doc, cur);
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "f"))
    _tlen_parse_file (tc, doc, cur);
  else
  if (!xmlStrcmp(cur->name, (const xmlChar *) "m"))
    _tlen_parse_notify (tc, doc, cur);
  else
  {
    tlen_debug ("Unknown tag.\n");
    xmlFreeDoc(doc);
    return FALSE;
  }

  xmlFreeDoc(doc);
  return TRUE;
}

static
void _tlen_parse_file (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  TlenEvent *ev;
  gchar *f, *n, *e, *i;
  gchar *hostname, *port;
  TlenP2p *p;


  tlen_debug ("Parsing file tag.\n");
  f = (gchar*) xmlGetProp(cur, "f");
  n = (gchar*) xmlGetProp(cur, "n");
  e = (gchar*) xmlGetProp(cur, "e");
  i = (gchar*) xmlGetProp(cur, "i");
  hostname = (gchar*) xmlGetProp(cur, "a");
  port = (gchar*) xmlGetProp(cur, "p");
  ev = tlen_event_new (TLEN_EVENT_FILE);
  p = _tlen_p2p_new (f);
  if (p)
  {
    p->i = i;
    p->filename = n;
  }
  TLEN_EVENT_FILE (ev)->p = p;
  if (!tlen_strcmp (e, "1"))
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_ASK;
  else
  if (!tlen_strcmp (e, "3"))
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_ERROR;
  else
  if (!tlen_strcmp (e, "4"))
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_DENIED;
  else
  if (!tlen_strcmp (e, "5"))
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_ACCEPTED;
  else
  if (!tlen_strcmp (e, "6"))
  {
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_OPENED;
    if (p)
    {
      p->hostname = hostname;
      p->port = atoi (port);
    }
  }
  else
  if (!tlen_strcmp (e, "7"))
  {
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_CONNECT;
    if (p)
    {
      p->hostname = hostname;
      p->port = atoi (port);
    }
  }
  else
    TLEN_EVENT_FILE (ev)->error = TLEN_P2P_UNKNOWN;
  g_free (f);
  g_free (e);
  g_free (port);
  tlen_event_add (tc, ev);
}

static
void _tlen_parse_notify (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  TlenEvent *ev;
  gchar *f;
  xmlChar *tp;
  TlenNotifyType t;

  tlen_debug ("Parsing notify tag.\n");
  f = (gchar*) xmlGetProp (cur, "f");
  tp = xmlGetProp (cur, "tp");

  if ((!xmlStrcmp(tp, (xmlChar *) "t")))
  {
    t = TLEN_NOTIFY_TYPING;
  } 
  else if ((!xmlStrcmp(tp, (xmlChar *) "u")))
  {
    t = TLEN_NOTIFY_STOPPED;
  }
  else// if ((!xmlStrcmp(tp, (xmlChar *) "a")))
  {
    t = TLEN_NOTIFY_ALERT;
  }

  ev = tlen_event_new (TLEN_EVENT_NOTIFY);
  TLEN_EVENT_NOTIFY (ev)->from = tlen_decode(f);
  TLEN_EVENT_NOTIFY (ev)->type = t;
  g_free (f);
  g_free ((gchar *) tp);
  tlen_event_add (tc, ev);
}

static
void _tlen_parse_webmsg (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  TlenEvent *ev;
  gchar *f, *e, *body;


  tlen_debug ("Parsing webmsg tag.\n");
  f = (gchar*) xmlGetProp(cur, "f");
  e = (gchar*) xmlGetProp(cur, "e");
  body = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
  ev = tlen_event_new (TLEN_EVENT_WEBMSG);
  TLEN_EVENT_WEBMSG(ev)->from = tlen_decode (f);
  TLEN_EVENT_WEBMSG(ev)->e = tlen_decode (e);
  TLEN_EVENT_WEBMSG(ev)->body = tlen_decode (body);
  g_free (f);
  g_free (e);
  g_free (body);
  tlen_event_add (tc, ev);
}

static
void _tlen_parse_newmail (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  TlenEvent *ev;
  gchar *f, *s;


  tlen_debug ("Parsing newmail tag.\n");
  f = (gchar*) xmlGetProp(cur, "f");
  s = (gchar*) xmlGetProp(cur, "s");
  ev = tlen_event_new (TLEN_EVENT_NEWMAIL);
  TLEN_EVENT_NEWMAIL(ev)->from = tlen_decode (f);
  TLEN_EVENT_NEWMAIL(ev)->e = tlen_decode (s);
  g_free (f);
  g_free (s);
  tlen_event_add (tc, ev);
}

static
void _tlen_parse_presence (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  xmlChar *type;
  TlenEvent *ev;
  TlenPresenceType status = TLEN_PRESENCE_AVAILABLE;
  gchar *p;

  tlen_debug ("Parsing presence tag.\n");
  type = xmlGetProp(cur, "type");
  key = xmlGetProp(cur, "from");
  ev = tlen_event_new (TLEN_EVENT_PRESENCE);
  p = strchr ((char*) key, '/');
  if (p)
    *p ='\0';
  TLEN_EVENT_PRESENCE (ev)->from = g_strdup ((gchar*) key);
  if (p)
    *p ='/';
  g_free ((gchar*) key);
  key = NULL;
  TLEN_EVENT_PRESENCE (ev)->type = g_strdup ((gchar*) type);
  cur = cur->xmlChildrenNode;
  while (cur != NULL)
  {
    if ((!xmlStrcmp (cur->name, (const xmlChar *) "show")))
    {
      key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
      if ((!xmlStrcmp (key, "available")))
        status = TLEN_PRESENCE_AVAILABLE;
      else
      if ((!xmlStrcmp (key, "away")))
        status = TLEN_PRESENCE_AWAY;
      else
      if ((!xmlStrcmp (key, "chat")))
        status = TLEN_PRESENCE_CHATTY;
      else
      if ((!xmlStrcmp (key, "xa")))
        status = TLEN_PRESENCE_EXT_AWAY;
      else
      if ((!xmlStrcmp (key, "dnd")))
        status = TLEN_PRESENCE_DND;
      else
        status = TLEN_PRESENCE_UNAVAILABLE;
      g_free ((gchar*)key);
    }
    else
    if ((!xmlStrcmp (cur->name, (const xmlChar *) "status")))
    {
      key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
      TLEN_EVENT_PRESENCE (ev)->description = tlen_decode ((gchar*) key);
      g_free ((gchar*)key);
    }
    cur = cur->next;
  }
  if ((!xmlStrcmp (type, "unavailable")))
  {
    status = TLEN_PRESENCE_UNAVAILABLE;
  }
  g_free ((gchar*) type);
  TLEN_EVENT_PRESENCE (ev)->status = status;
  tlen_event_add (tc, ev);
  return;
}

static
void _tlen_parse_s (TlenConnection *tc, xmlNodePtr cur)
{
  xmlChar *sid;

  tlen_debug ("Parsing s tag.\n");
  sid = xmlGetProp(cur, "i");
  tlen_debug ("Got id: %s\n", sid);
  tc->sid = (gchar*) sid;
  _tlen_auth_send (tc);
  return;
}

static
void _tlen_parse_iq (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *type, *id, *from;
  TlenEvent *ev;

  tlen_debug ("Parsing iq tag.\n");
  type = xmlGetProp(cur, "type");
  if (type)
  {
    tlen_debug ("Iq type: %s\n", type);
    if (!xmlStrcmp (type, "error"))
    {
      if (tc->state == TLEN_STATE_LOGGING)
      {
        ev = tlen_event_new (TLEN_EVENT_UNAUTHORIZED);
        tlen_event_add (tc, ev);
      }
      else
        tlen_debug ("Unknown error.\n");
    }
    else
    if (!xmlStrcmp (type, "result"))
    {
      if (tc->state == TLEN_STATE_LOGGING)
      {
        tc->state = TLEN_STATE_LOGGED;
        g_timeout_add(55000, _tlen_ping, tc);
        ev = tlen_event_new (TLEN_EVENT_LOGGED);
        tlen_event_add (tc, ev);
      }
      else
        if (tc->state == TLEN_STATE_LOGGED)
        {
          id = xmlGetProp(cur, "id");
          from = xmlGetProp(cur, "from");
          if (id)
          {
            if (!xmlStrcmp (id, "GetRoster"))
            {
              _tlen_parse_roster (tc, doc,  cur);
            }
            else
            if (!xmlStrcmp (id, "tr"))
            {
              if (!xmlStrcmp (from, "tuba"))
              {
                TlenPubdirItem *item;

                if (cur)
                  cur = cur->xmlChildrenNode;
                if (cur)
                  cur = cur->xmlChildrenNode;
                item = _tlen_parse_tuba_item (tc, doc, cur);
                ev = tlen_event_new (TLEN_EVENT_PUBDIR);
                TLEN_EVENT_PUBDIR (ev)->item = item;
                tlen_event_add (tc, ev);
              }
            }
            else
            if (!xmlStrcmp (id, "tw"))
            {
              if (!xmlStrcmp (from, "tuba"))
              {
                ev = tlen_event_new (TLEN_EVENT_PUBDIR_CHANGED);
                tlen_event_add (tc, ev);
              }
            }
            else
            if (!xmlStrcmp (id, "src"))
            {
              if (!xmlStrcmp (from, "tuba"))
                _tlen_parse_search_result (tc, doc, cur);
            }
            g_free ((gchar*) id);
            g_free ((gchar*) from);
          }
        }
        else
          tlen_debug ("Unknown iq result.\n");
    }
    else
    if (!xmlStrcmp (type, "set"))
    {
      xmlNodePtr q = cur->xmlChildrenNode;

      if (q && q->ns)
      {
        if (!xmlStrcmp (q->ns->href, "jabber:iq:roster"))
        {
          _tlen_parse_roster (tc, doc,  cur);
        }
      }
    }
    else
      tlen_debug ("Unknown iq type.\n");
    g_free ((gchar*)type);
  }
  else
    tlen_debug ("Unknown iq type.\n");
  return;
}

static
void _tlen_parse_search_result (TlenConnection *tc, xmlDocPtr doc,
                                xmlNodePtr cur)
{
  TlenEvent *ev;
  TlenPubdirItem *item;

  if (cur)
    cur = cur->xmlChildrenNode;
  if (cur)
    cur = cur->xmlChildrenNode;
  ev = tlen_event_new (TLEN_EVENT_SEARCH);
  while (cur)
  {
    item = _tlen_parse_tuba_item (tc, doc, cur);
    TLEN_EVENT_SEARCH(ev)->list = g_slist_append (TLEN_EVENT_SEARCH(ev)->list,
                                                   item);
    cur = cur->next;
  }
  tlen_event_add (tc, ev);
}

static
TlenPubdirItem *_tlen_parse_tuba_item (TlenConnection *tc, xmlDocPtr doc,
                                       xmlNodePtr cur)
{
  TlenPubdirItem *item = NULL;;
  gchar *s;

  if (!cur)
  {
    tlen_debug ("Empty tuba result.\n");
    item = tlen_pubdir_item_new ();
    return item;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *)"item"))
    return NULL;
  tlen_debug ("Parsing tuba result.\n");
  item = tlen_pubdir_item_new ();
  item->id = (gchar*) xmlGetProp(cur, "jid");
  if (cur)
    cur = cur->xmlChildrenNode;
  while (cur)
  {
    s = (gchar*) xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"first")))
      item->firstname = tlen_decode (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"last")))
      item->lastname = tlen_decode (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"nick")))
      item->nick = tlen_decode (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"email")))
      item->email = tlen_decode (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"e")))
      item->school = tlen_decode (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"c")))
      item->city = tlen_decode (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"s")))
      item->gender = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"r")))
      item->look_for = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"b")))
      item->age = item->birthyear = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"g")))
      item->voice = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"j")))
      item->job = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"p")))
      item->plans = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"v")))
      item->visible = atoi (s);
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"a")))
    {
      item->status = atoi (s);
    }
    g_free (s);
    cur = cur->next;
  }
  return item;
}

static
void _tlen_parse_roster (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  gchar *s;
  TlenRosterItem *item;
  TlenEvent *ev;
  xmlNodePtr gr;

  tlen_debug ("Parsing roster.\n");
  if (cur)
    cur = cur->xmlChildrenNode;
  if (cur)
    cur = cur->xmlChildrenNode;
  if (!cur)
    return;
  ev = tlen_event_new (TLEN_EVENT_ROSTER);
  if (!ev)
    return;
  while (cur)
  {
    if ((!xmlStrcmp(cur->name, (const xmlChar *)"item")))
    {
      item = g_new0 (TlenRosterItem, 1);
      if (item)
      {
        s = (gchar*) xmlGetProp (cur, "jid");
        item->jid = s;
        s = (gchar*) xmlGetProp (cur, "name");
        item->name = tlen_decode((gchar*) s);
        s = (gchar*) xmlGetProp (cur, "subscription");
        item->subscription = s;
        s = (gchar*) xmlGetProp (cur, "ask");
        item->ask = s;
        gr = cur->xmlChildrenNode;

        while (gr)
        {
          if ((!xmlStrcmp(gr->name, (const xmlChar *)"group")))
          {
            tlen_debug ("grupa\n");
            s = (gchar*) xmlNodeListGetString(doc, gr->xmlChildrenNode, 1);
            item->group = tlen_decode (s);
            g_free (s);
            break;
          }
          gr = gr->next;
        }
        TLEN_EVENT_ROSTER (ev)->list = g_slist_append(TLEN_EVENT_ROSTER (ev)->list, (gpointer) item);
      }
    }
    cur = cur->next;
  }
  tlen_event_add (tc, ev);
}

static
void _tlen_parse_message (TlenConnection *tc, xmlDocPtr doc, xmlNodePtr cur)
{
  gchar *s;
  TlenEvent *ev;
  /* message vars */
  TlenMessageType msgtype;
  gchar *msgfrom, *msgbody, *msgstamp;
  /* advert vars */
//  gchar *adname, *adurl;
//  gint adwidth, adheight;

  /* Parse whole message tag */
  msgfrom = (gchar*) xmlGetProp (cur, "from");
  msgbody = msgstamp = NULL;
  s = (gchar*) xmlGetProp (cur, "type");
  if (s)
  {
    if (!tlen_strcmp (s, "normal"))
      msgtype = TLEN_MESSAGE_TYPE_NORMAL;
    else
      msgtype = TLEN_MESSAGE_TYPE_CHAT;
    g_free (s);
  }
  else
  {
    msgtype = TLEN_MESSAGE_TYPE_CHAT;
  }

  cur = cur->xmlChildrenNode;
  while (cur)
  {
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "body")))
    {
      s = (gchar*) xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
      if (s)
      {
        msgbody = tlen_decode (s);
        g_free (s);
      }
    }
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "x")))
    {
      msgstamp = (gchar*) xmlGetProp (cur, "stamp");
    }
    else
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "no")))
    {
      tlen_debug ("Adverb, ignoring...\n");
      if (msgbody)
        g_free (msgbody);
      if (msgfrom)
        g_free (msgfrom);
      return;
    }
    cur = cur->next;
  }

  tlen_debug("Make message event, and set all variables");
  ev = tlen_event_new (TLEN_EVENT_MESSAGE);
  TLEN_EVENT_MESSAGE (ev)->from =  msgfrom;
  TLEN_EVENT_MESSAGE (ev)->stamp = msgstamp;
  TLEN_EVENT_MESSAGE (ev)->body = msgbody;
  TLEN_EVENT_MESSAGE (ev)->type = msgtype;

  tlen_event_add (tc, ev);
}

