/* Libtlen2 - multiplatform Tlen.pl library
 * Copyright (C) 2003  Libtlen Team
 *
 * $Id: libtlen2-utils.c,v 1.5 2003/12/30 18:09:42 who_ami Exp $
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include "libtlen2-utils.h"
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

/* debug level */
static guint debug_level;

/**
  * tlen_decode:
  * @what: string to decode.
  *
  * Decodes @what from URLencode.
  *
  * returns: newly allocated string.
  **/
gchar *tlen_decode(const gchar *what)
{
  guchar *s = (guchar*) what;
  GString *str;

  if (what)
  {
    str = g_string_new ("");
    if (!str)
      return NULL;

    while (*s)
    {
      if (*s == '+')
        str = g_string_append_c (str, ' ');
      else if ((*s == '%') && s[1] && s[2] && isxdigit ((int) s[1])
               && isxdigit ((int) s[2]))
      {
        int code;
        sscanf (s + 1, "%2x", &code);
        if (code != '\r')
          str = g_string_append_c (str, (guchar) code);
        s += 2;
      }
      else
        str = g_string_append_c (str, *s);
      s++;
    }
    return g_string_free (str, FALSE);
  }
  return NULL;
}

/**
  * tlen_encode:
  * @what: string to encode.
  *
  * Encodes @what using URLencode.
  *
  * returns: newly allocated string.
  **/
gchar *tlen_encode (const gchar *what)
{
  guchar *s = (guchar*) what;
  GString *str;

  if (what)
  {
    str = g_string_new ("");
    if (!str)
      return NULL;
    while (*s)
    {
      if (*s == ' ')
        str = g_string_append_c (str, '+');
      else
        if ((*s < '0' && *s != '-' && *s != '.')
            || (*s < 'A' && *s > '9') || (*s > 'Z'
                                          && *s < 'a'
                                          && *s != '_')
            || (*s > 'z'))
        {
          g_string_append_printf (str, "%%%02X", *s);
        }
        else
          str = g_string_append_c (str, *s);
      s++;
    }
    return g_string_free (str, FALSE);
  }
  return NULL;
}



/**
  * tlen_debug_set:
  * @level: debug level.
  *
  * Sets debug level to @level.
  **/
void
tlen_debug_set(guint level)
{
  debug_level = level;
}

/**
  * tlen_debug_get:
  *
  * Gets debug level.
  **/
guint
tlen_debug_get(void)
{
  return debug_level;
}

/* API ends here */

void
_tlen_debug(int level, const gchar *name, gchar *fmt, ...)
{
  va_list ap;
  int d;
  char c, *s;

  switch (level)
  {
    case TLEN_DEBUG_NORMAL:
      g_print ("--- %s ---\n", name);
      break;
    case TLEN_DEBUG_HIGH:
      g_print ("+++ %s +++\n", name);
      break;
  }

  va_start(ap, fmt);
  while (*fmt)
  {
    switch (*fmt)
    {
      case '%':
        if (*++fmt)
          switch (*fmt) {
            case 's':           /* string */
              s = va_arg(ap, char *);
              g_print("%s", s);
              break;
            case 'd':           /* int */
              d = va_arg(ap, int);
              g_print("%d", d);
              break;
            case 'c':           /* char */
              /* need a cast here since va_arg only
                 takes fully promoted types */
              c = (char) va_arg(ap, int);
              g_print("%c", c);
              break;
          }
        break;
      case '\\':
        if (*++fmt)
          switch (*fmt) {
            case 'n': /* new line */
              g_print ("\n");
              break;
          }
        break;
      default:
        g_print("%c", *fmt);
    }
    *fmt++;
  }
  va_end(ap);
}

/**
  * tlen_strlen:
  * @s: ascii string.
  *
  * Computes lenght of the string @s.
  *
  * returns: string lenght. The function returns 0 if the string is NULL.
  **/
guint tlen_strlen (const gchar *s)
{
  if (!s)
    return 0;
  return strlen (s);
}

gint tlen_strcmp (const gchar *s1, const gchar *s2)
{
  if (!s1 && !s2)
    return 0;
  if (!s1 || !s2)
    return 1;
  return strcmp (s1, s2);
}
